import textwrap

from r2pb.parser import parse_msg_content
from r2pb.generator import ProtoGenerator


def test_generate_proto_with_dependencies():
    """Test generating a .proto file with dependencies."""
    msg_content = textwrap.dedent("""
        # Test message
        string name
        std_msgs/Header header
        another_pkg/AnotherMessage data
        """)
    parsed_msg = parse_msg_content(msg_content)
    generator = ProtoGenerator()

    proto_content, dependencies = generator.generate_proto(
        parsed_msg,
        package_name="my_package",
        msg_name="Person"
    )

    expected_proto = textwrap.dedent("""
        // Generated by r2pb - from my_package/Person.msg
        syntax = "proto3";

        package my_package;

        import "another_pkg/AnotherMessage.proto";
        import "std_msgs/Header.proto";

        message Person {
          string name = 1;
          std_msgs.Header header = 2;
          another_pkg.AnotherMessage data = 3;
        }
        """).strip()

    actual_lines = [line.strip() for line in proto_content.strip().splitlines() if line.strip()]
    expected_lines = [line.strip() for line in expected_proto.strip().splitlines() if line.strip()]

    assert actual_lines == expected_lines
    assert sorted(dependencies) == ["another_pkg/AnotherMessage", "std_msgs/Header"]


def test_generate_proto_no_dependencies():
    """Test generating a .proto file with no dependencies."""
    msg_content = textwrap.dedent("""
        # Simple message
        string first_name
        int32 age
        """)
    parsed_msg = parse_msg_content(msg_content)
    generator = ProtoGenerator()

    proto_content, dependencies = generator.generate_proto(
        parsed_msg,
        package_name="my_package",
        msg_name="User"
    )

    expected_proto = textwrap.dedent("""
        // Generated by r2pb - from my_package/User.msg
        syntax = "proto3";

        package my_package;

        message User {
          string first_name = 1;
          int32 age = 2;
        }
        """).strip()

    actual_lines = [line.strip() for line in proto_content.strip().splitlines() if line.strip()]
    expected_lines = [line.strip() for line in expected_proto.strip().splitlines() if line.strip()]

    assert actual_lines == expected_lines
    assert dependencies == []


import argparse
import sys
from .converter import Converter

def main():
    """Main function for the r2pb command-line interface."""
    parser = argparse.ArgumentParser(
        description='r2pb: ROS .msg to Protobuf .proto converter.'
    )
    parser.add_argument(
        'msg_type',
        type=str,
        help='The ROS message type to convert (e.g., std_msgs/String).'
    )
    parser.add_argument(
        '-o', '--output-dir',
        type=str,
        default='.',
        help='The directory where the .proto files will be saved.'
    )
    parser.add_argument(
        '-d', '--ros-distro',
        type=str,
        default='noetic',
        help='The ROS distribution to use (e.g., noetic, melodic).'
    )

    args = parser.parse_args()

    print(f"Converting {args.msg_type} for ROS {args.ros_distro}...")
    print(f"Output directory: {args.output_dir}")

    try:
        converter = Converter(ros_distro=args.ros_distro)
        converter.convert(args.msg_type, args.output_dir)
        print("\nConversion finished successfully.")
    except Exception as e:
        print(f"\nAn error occurred during conversion: {e}", file=sys.stderr)
        print("Conversion failed.", file=sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()


import re
from pathlib import Path
from typing import List, Dict, Any, Tuple, NamedTuple, Optional, Union

from .fetcher import RosMsgFetcher


class Field(NamedTuple):
    field_type: str
    name: str

class Constant(NamedTuple):
    const_type: str
    name: str
    value: Any

class ParsedMsg(NamedTuple):
    fields: List[Field]
    constants: List[Constant]

def parse_msg_content(content: str) -> ParsedMsg:
    """Parses the content of a .msg file into a structured format."""
    fields = []
    constants = []

    lines = content.split('\n')

    for line in lines:
        line = line.strip()
        # Ignore comments and empty lines
        if not line or line.startswith('#'):
            continue

        # Split comments from the definition
        if '#' in line:
            line = line.split('#', 1)[0].strip()

        # Check for constants
        if '=' in line:
            parts = [p.strip() for p in line.split('=', 1)]
            const_def, const_val = parts
            const_type, const_name = const_def.split()
            constants.append(Constant(const_type, const_name, const_val))
        # Check for fields
        else:
            parts = line.split()
            if len(parts) >= 2:
                field_type, field_name = parts[:2]
                fields.append(Field(field_type, field_name))

    return ParsedMsg(fields=fields, constants=constants)


class MsgParser:
    """ROS 消息文件解析器，支持本地搜索和在线获取。"""

    def __init__(self, local_package_paths: Optional[List[Union[str, Path]]] = None):
        self.local_package_paths = [Path(p) for p in local_package_paths] if local_package_paths else []
        self.fetcher = RosMsgFetcher()

    def find_msg_file_content(self, package_name: str, msg_name: str) -> str:
        """查找指定的消息文件内容，优先在本地搜索，找不到则尝试在线获取。

        Args:
            package_name: ROS 包名，如 'std_msgs'。
            msg_name: 消息名称，如 'String' (不带 .msg 后缀)。

        Returns:
            消息文件的文本内容。

        Raises:
            FileNotFoundError: 当消息文件无法在本地和在线仓库中找到时。
        """
        # 1. 先在本地包路径中查找
        content = self._find_local_msg_content(package_name, msg_name)
        if content is not None:
            return content

        # 2. 本地找不到，尝试在线获取
        try:
            package_path = self.fetcher.find_and_fetch(package_name)
            msg_file_path = self._find_msg_in_package(package_path, msg_name)
            if msg_file_path:
                return msg_file_path.read_text(encoding='utf-8')
        except (KeyError, FileNotFoundError):
            # 捕获 fetcher 找不到包或文件找不到的异常，统一处理
            pass

        raise FileNotFoundError(
            f"Message '{msg_name}.msg' not found in package '{package_name}' "
            f"(searched in {self.local_package_paths} and online)"
        )

    def _find_local_msg_content(self, package_name: str, msg_name: str) -> Optional[str]:
        """在配置的本地路径中查找消息文件并返回其内容。"""
        for base_path in self.local_package_paths:
            # 假设的目录结构: base_path / package_name / msg / msg_name.msg
            msg_file = base_path / package_name / 'msg' / f'{msg_name}.msg'
            if msg_file.is_file():
                return msg_file.read_text(encoding='utf-8')
        return None

    def _find_msg_in_package(self, package_path: Path, msg_name: str) -> Optional[Path]:
        """在给定的包路径下查找 .msg 文件。"""
        # ROS 标准结构是在包目录下的 'msg' 子目录中
        msg_file = package_path / 'msg' / f'{msg_name}.msg'
        if msg_file.is_file():
            return msg_file
        
        # 有些包可能直接把 .msg 文件放在根目录
        msg_file_root = package_path / f'{msg_name}.msg'
        if msg_file_root.is_file():
            return msg_file_root
            
        return None

    def parse(self, package_name: str, msg_name: str) -> ParsedMsg:
        """查找并解析一个消息文件。

        这是推荐使用的主方法，它封装了查找和解析的整个过程。
        """
        content = self.find_msg_file_content(package_name, msg_name)
        return parse_msg_content(content)


import pytest
from r2pb.mapper import map_ros_to_proto_type


@pytest.mark.parametrize(
    "ros_type, proto_type",
    [
        ("string", "string"),
        ("int32", "int32"),
        ("uint64", "uint64"),
        ("float32", "float"),
        ("time", "google.protobuf.Timestamp"),
        ("std_msgs/Header", "Header"),
        ("geometry_msgs/Pose", "Pose"),
    ],
)
def test_map_ros_to_proto_type(ros_type, proto_type):
    """Tests mapping of various ROS types to Protobuf types."""
    assert map_ros_to_proto_type(ros_type) == proto_type


# A mapping from ROS built-in types to Protobuf types
ROS_TO_PROTO_TYPE_MAP = {
    'bool': 'bool',
    'byte': 'int32',  # Protobuf doesn't have a byte type, int32 is a common choice
    'char': 'string', # Protobuf doesn't have a char type, string is a common choice
    'float32': 'float',
    'float64': 'double',
    'int8': 'int32',
    'uint8': 'uint32',
    'int16': 'int32',
    'uint16': 'uint32',
    'int32': 'int32',
    'uint32': 'uint32',
    'int64': 'int64',
    'uint64': 'uint64',
    'string': 'string',
    # Special ROS types
    'time': 'google.protobuf.Timestamp',
    'duration': 'google.protobuf.Duration',
}

def map_ros_to_proto_type(ros_type: str) -> str:
    """Maps a ROS type to its corresponding Protobuf type."""
    if ros_type in ROS_TO_PROTO_TYPE_MAP:
        return ROS_TO_PROTO_TYPE_MAP[ros_type]
    # For complex types (e.g., other messages), we assume they will be
    # converted to a message of the same name in the same package.
    # We will handle package path resolution later.
    return ros_type.split('/')[-1]


import pytest
from pathlib import Path
from unittest import mock

from r2pb.converter import Converter
from r2pb.parser import ParsedMsg, Field


def test_converter_basic(tmp_path: Path):
    """Test converting a single message with dependencies."""
    # 1. Setup: Mock parser and generator
    mock_parser = mock.Mock()
    mock_generator = mock.Mock()

    # Mock data for 'my_pkg/MyMessage'
    my_message_parsed = ParsedMsg(fields=[Field(name='header', field_type='std_msgs/Header')], constants=[])
    my_message_proto = "syntax = \"proto3\"; package my_pkg; import \"std_msgs/Header.proto\"; message MyMessage { std_msgs.Header header = 1; }", ["std_msgs/Header"]

    # Mock data for 'std_msgs/Header' based on real definition
    header_parsed = ParsedMsg(
        fields=[
            Field(name='seq', field_type='uint32'),
            Field(name='stamp', field_type='time'),
            Field(name='frame_id', field_type='string'),
        ],
        constants=[]
    )
    header_proto = 'syntax = "proto3"; package std_msgs; import "google/protobuf/timestamp.proto"; message Header { uint32 seq = 1; google.protobuf.Timestamp stamp = 2; string frame_id = 3; }', []

    # Configure mocks
    mock_parser.parse.side_effect = lambda pkg, msg: \
        my_message_parsed if (pkg, msg) == ('my_pkg', 'MyMessage') else \
        header_parsed if (pkg, msg) == ('std_msgs', 'Header') else \
        None

    mock_generator.generate_proto.side_effect = lambda parsed, **kwargs: \
        my_message_proto if parsed == my_message_parsed else \
        header_proto if parsed == header_parsed else \
        ("", [])

    # 2. Execution: Run the converter
    output_dir = tmp_path / "proto_out"
    converter = Converter()
    converter._parser = mock_parser
    converter._generator = mock_generator

    converter.convert("my_pkg/MyMessage", output_dir)

    # 3. Verification
    # Check that parser and generator were called correctly
    assert mock_parser.parse.call_count == 2
    mock_parser.parse.assert_any_call('my_pkg', 'MyMessage')
    mock_parser.parse.assert_any_call('std_msgs', 'Header')

    assert mock_generator.generate_proto.call_count == 2

    # Check that files were written correctly
    my_message_file = output_dir / 'my_pkg' / 'MyMessage.proto'
    header_file = output_dir / 'std_msgs' / 'Header.proto'
    assert my_message_file.exists()
    assert header_file.exists()
    assert my_message_file.read_text() == my_message_proto[0]
    assert header_file.read_text() == header_proto[0]


def test_converter_failure(tmp_path: Path, capsys):
    """Test that the converter handles failures gracefully."""
    mock_parser = mock.Mock()
    mock_generator = mock.Mock()

    # 'bad_pkg/BadMessage' will fail to parse
    mock_parser.parse.side_effect = ValueError("File not found")

    output_dir = tmp_path / "proto_out"
    converter = Converter()
    converter._parser = mock_parser
    converter._generator = mock_generator

    # Run conversion, expecting it to fail
    with pytest.raises(ValueError, match="File not found"):
        converter.convert("bad_pkg/BadMessage", output_dir)

    # Check that an error message was printed and no files were created
    captured = capsys.readouterr()
    assert "Failed to convert bad_pkg/BadMessage" in captured.out
    assert not (output_dir / 'bad_pkg').exists()


from pathlib import Path
from typing import List, Dict, Any
from jinja2 import Environment, FileSystemLoader
from .parser import ParsedMsg, Field, Constant, MsgParser # 引入 MsgParser
from .mapper import map_ros_to_proto_type


class ProtoField:
    """Represents a field in the Protobuf message."""
    def __init__(self, name: str, proto_type: str, package: str = ''):
        self.name = name
        self.proto_type = proto_type
        self.package = package


class ProtoGenerator:
    """Generates .proto files from ROS message definitions."""
    def __init__(self):
        templates_dir = Path(__file__).parent / 'templates'
        self.env = Environment(
            loader=FileSystemLoader(str(templates_dir)),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.template = self.env.get_template('msg.proto.j2')

    def _collect_dependencies(self, fields: List[Field]) -> List[str]:
        """Collects required dependencies based on field types."""
        dependencies = set()
        for field in fields:
            if '/' in field.field_type:
                dependencies.add(field.field_type)
        return sorted(list(dependencies))

    def _convert_fields(self, fields: List[Field]) -> List[ProtoField]:
        """Converts ROS message fields to Protobuf fields."""
        proto_fields = []
        for field in fields:
            package = ''
            ros_type = field.field_type
            if '/' in ros_type:
                package, ros_type = ros_type.split('/', 1)

            proto_fields.append(
                ProtoField(
                    name=field.name,
                    proto_type=map_ros_to_proto_type(field.field_type),
                    package=package
                )
            )
        return proto_fields

    def generate_proto(self, parsed_msg: ParsedMsg, package_name: str, msg_name: str) -> (str, List[str]):
        """Generates a .proto file content for a given ROS message."""
        proto_fields = self._convert_fields(parsed_msg.fields)
        dependencies = self._collect_dependencies(parsed_msg.fields)
        
        # 从依赖项生成导入语句
        imports = sorted([f'{dep}.proto' for dep in dependencies])

        proto_content = self.template.render(
            package_name=package_name,
            msg_name=msg_name,
            fields=proto_fields,
            constants=parsed_msg.constants,
            imports=imports,
        )
        return proto_content, dependencies


import pytest
from pathlib import Path
from unittest.mock import MagicMock, patch

from r2pb.parser import parse_msg_content, ParsedMsg, Field, Constant, MsgParser

# --- Tests for the original parse_msg_content function ---

def test_parse_simple_msg():
    """Tests parsing of a simple .msg file content."""
    content = """
# This is a comment
string name
int32 age

# Constants
string FOO = bar
int32 BAZ = 123

"""
    expected = ParsedMsg(
        fields=[
            Field(field_type='string', name='name'),  # 修改这里：type -> field_type
            Field(field_type='int32', name='age'),    # 修改这里：type -> field_type
        ],
        constants=[
            Constant(const_type='string', name='FOO', value='bar'),  # 注意：这里也要用 const_type
            Constant(const_type='int32', name='BAZ', value='123'),   # 注意：这里也要用 const_type
        ]
    )
    assert parse_msg_content(content) == expected


def test_parse_complex_msg():
    """Tests parsing of a .msg file with complex types."""
    content = """
# A complex message
std_msgs/Header header
geometry_msgs/Pose pose
string child_frame_id
"""
    expected = ParsedMsg(
        fields=[
            Field(field_type='std_msgs/Header', name='header'),
            Field(field_type='geometry_msgs/Pose', name='pose'),
            Field(field_type='string', name='child_frame_id'),
        ],
        constants=[]
    )
    assert parse_msg_content(content) == expected


# --- Tests for the new MsgParser class ---

@pytest.fixture
def mock_fetcher(monkeypatch):
    """Mock the RosMsgFetcher to simulate online fetching."""
    mock_fetcher_instance = MagicMock()
    # Use patch to replace the class in the parser module
    with patch('r2pb.parser.RosMsgFetcher', return_value=mock_fetcher_instance) as mock_class:
        yield mock_fetcher_instance

@pytest.fixture
def local_packages(tmp_path):
    """Create a dummy local package structure."""
    local_path = tmp_path / "ros_ws"
    local_path.mkdir()
    
    # Create a package: my_msgs
    my_msgs_path = local_path / "my_msgs" / "msg"
    my_msgs_path.mkdir(parents=True)
    (my_msgs_path / "MyData.msg").write_text("string data")
    
    return local_path

def test_parser_find_local_file(local_packages):
    """Test finding a message file from a local path."""
    parser = MsgParser(local_package_paths=[local_packages])
    content = parser.find_msg_file_content("my_msgs", "MyData")
    assert content == "string data"

def test_parser_local_file_not_found(local_packages):
    """Test that FileNotFoundError is raised if a local file is not found."""
    parser = MsgParser(local_package_paths=[local_packages])
    with pytest.raises(FileNotFoundError):
        # This should fail because we don't mock the fetcher, so online lookup also fails
        parser.find_msg_file_content("my_msgs", "NonExistent")

def test_parser_fetch_online_when_not_local(mock_fetcher, tmp_path):
    """Test that the parser attempts to fetch online if not found locally."""
    # Simulate the fetcher returning a path to a 'downloaded' package
    online_pkg_path = tmp_path / "online_pkgs" / "std_msgs"
    (online_pkg_path / "msg").mkdir(parents=True)
    (online_pkg_path / "msg" / "String.msg").write_text("string data")
    mock_fetcher.find_and_fetch.return_value = online_pkg_path

    # Initialize parser with no local paths
    parser = MsgParser(local_package_paths=[])
    content = parser.find_msg_file_content("std_msgs", "String")

    # Verify that fetcher was called and content is correct
    mock_fetcher.find_and_fetch.assert_called_once_with("std_msgs")
    assert content == "string data"

def test_parser_fetch_online_fails(mock_fetcher):
    """Test that FileNotFoundError is raised when online fetch fails."""
    # Simulate a failure in the fetcher
    mock_fetcher.find_and_fetch.side_effect = KeyError("Package not found")

    parser = MsgParser()
    with pytest.raises(FileNotFoundError, match="not found in package 'non_existent_pkg'"):
        parser.find_msg_file_content("non_existent_pkg", "SomeMsg")

def test_parser_parse_method(mock_fetcher, tmp_path):
    """Test the main 'parse' method which integrates finding and parsing."""
    # Setup a fetched file
    online_pkg_path = tmp_path / "online_pkgs" / "sensor_msgs"
    (online_pkg_path / "msg").mkdir(parents=True)
    (online_pkg_path / "msg" / "Imu.msg").write_text("std_msgs/Header header\nint32 data")
    mock_fetcher.find_and_fetch.return_value = online_pkg_path

    parser = MsgParser()
    parsed_result = parser.parse("sensor_msgs", "Imu")

    expected = ParsedMsg(
        fields=[
            Field(field_type='std_msgs/Header', name='header'),
            Field(field_type='int32', name='data'),
        ],
        constants=[]
    )
    assert parsed_result == expected


def test_parser_online_repo_exists_but_msg_is_missing(mock_fetcher, tmp_path):
    """测试当远程仓库成功获取，但里面没有所需的 .msg 文件时的情况。"""
    # 1. 模拟 fetcher 成功返回了一个包的路径
    online_pkg_path = tmp_path / "online_pkgs" / "std_msgs"
    online_pkg_path.mkdir(parents=True)
    # 2. **但是**，我们故意不在这个路径下创建 'String.msg' 文件
    
    mock_fetcher.find_and_fetch.return_value = online_pkg_path

    parser = MsgParser()

    # 3. 断言：即使仓库找到了，但因为最终找不到文件，仍然应该抛出 FileNotFoundError
    with pytest.raises(FileNotFoundError, match="'String.msg' not found in package 'std_msgs'"):
        parser.find_msg_file_content("std_msgs", "String")

    # 验证 fetcher 确实被调用了
    mock_fetcher.find_and_fetch.assert_called_once_with("std_msgs")


import unittest
from unittest.mock import patch, MagicMock
from r2pb import cli

class TestCli(unittest.TestCase):

    @patch('r2pb.cli.Converter')
    @patch('r2pb.cli.argparse.ArgumentParser.parse_args')
    def test_main_flow(self, mock_parse_args, mock_converter_class):
        """Test the main CLI flow with mocked arguments and converter."""
        # Arrange: Mock the parsed arguments
        mock_args = MagicMock()
        mock_args.msg_type = 'std_msgs/String'
        mock_args.output_dir = '/tmp/proto_test'
        mock_args.ros_distro = 'noetic'
        mock_parse_args.return_value = mock_args

        # Arrange: Mock the Converter instance and its methods
        mock_converter_instance = MagicMock()
        mock_converter_class.return_value = mock_converter_instance

        # Act: Call the main function
        cli.main()

        # Assert: Check if Converter was initialized and called correctly
        mock_converter_class.assert_called_once_with(ros_distro='noetic')
        mock_converter_instance.convert.assert_called_once_with(
            'std_msgs/String', '/tmp/proto_test'
        )

if __name__ == '__main__':
    unittest.main()


import os
import shutil
import tempfile
from pathlib import Path
from git import Repo, GitCommandError

# 预设的 ROS 消息仓库
ROS_MSG_REPOS = {
    "common_msgs": "https://github.com/ros/common_msgs.git",
    "std_msgs": "https://github.com/ros/std_msgs.git",
    # ... 可以根据需要添加更多仓库
}

class RosMsgFetcher:
    """从远程 Git 仓库获取并缓存 ROS 消息包。"""

    def __init__(self, cache_dir: Path = None):
        if cache_dir is None:
            self.cache_dir = Path.home() / '.cache' / 'r2pb'
        else:
            self.cache_dir = cache_dir
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def fetch_package(self, package_name: str, repo_url: str) -> Path:
        """
        从指定的 Git 仓库 URL 下载或更新一个 ROS 包。

        Args:
            package_name: 要获取的包名 (例如 'sensor_msgs')。
            repo_url: 包含该包的 Git 仓库 URL。

        Returns:
            包在本地缓存中的路径。
        """
        repo_name = Path(repo_url).stem
        repo_path = self.cache_dir / repo_name

        try:
            if repo_path.exists():
                print(f"Updating repository: {repo_name}...")
                repo = Repo(repo_path)
                repo.remotes.origin.pull()
            else:
                print(f"Cloning repository: {repo_name} from {repo_url}...")
                Repo.clone_from(repo_url, repo_path)
        except GitCommandError as e:
            print(f"Error fetching repository {repo_url}: {e}")
            raise

        # Check if a subdirectory with the package name exists.
        # If not, assume the repo root is the package path (e.g., for std_msgs repo).
        package_path = repo_path / package_name
        if not package_path.is_dir():
            # If the subdirectory doesn't exist, maybe the repo itself is the package.
            if (repo_path / 'package.xml').exists() or (repo_path / 'msg').is_dir():
                return repo_path
            else:
                raise FileNotFoundError(f"Package '{package_name}' not found in repository '{repo_name}'.")

        return package_path

    def find_and_fetch(self, package_name: str) -> Path:
        """
        在预设的仓库中查找并获取一个 ROS 包。

        Args:
            package_name: 要查找的包名。

        Returns:
            包在本地缓存中的路径。

        Raises:
            KeyError: 如果在任何预设的仓库中都找不到该包。
        """
        # 简单的实现：假设包名直接对应仓库名或在 common_msgs 中
        # 更复杂的实现可以查询一个清单文件
        if package_name in ROS_MSG_REPOS:
            repo_url = ROS_MSG_REPOS[package_name]
            return self.fetch_package(package_name, repo_url)
        
        # 尝试在 common_msgs 中查找
        common_msgs_url = ROS_MSG_REPOS.get("common_msgs")
        if common_msgs_url:
            try:
                return self.fetch_package(package_name, common_msgs_url)
            except FileNotFoundError:
                pass # 在 common_msgs 中没找到，继续

        raise KeyError(f"Package '{package_name}' not found in any known repository.")


import pytest
from pathlib import Path
from unittest.mock import MagicMock

from r2pb.fetcher import RosMsgFetcher, ROS_MSG_REPOS


@pytest.fixture
def mock_git_repo(monkeypatch):
    """模拟 git.Repo 类及其 clone_from 方法。"""
    mock_repo_instance = MagicMock()
    mock_repo_class = MagicMock(return_value=mock_repo_instance)

    def mock_clone_from_side_effect(url, path):
        repo_dir = Path(path)
        # 模拟 `git clone`: 创建仓库根目录
        repo_dir.mkdir(parents=True, exist_ok=True)

        # 在仓库内创建模拟的包目录
        repo_name = Path(url).stem
        if repo_name == "common_msgs":
            (repo_dir / "sensor_msgs").mkdir()
        elif repo_name == "std_msgs":
            (repo_dir / "std_msgs").mkdir()
        
        # clone_from 应该返回一个 Repo 实例
        return mock_repo_instance

    # 1. 使用 monkeypatch 替换真实的的 Repo 类为一个 Mock 类
    monkeypatch.setattr("r2pb.fetcher.Repo", mock_repo_class)
    
    # 2. 配置 Mock 类的 clone_from 方法的行为
    #    它现在是一个具有 side_effect 的 MagicMock，可以被断言
    mock_repo_class.clone_from.side_effect = mock_clone_from_side_effect
    
    return mock_repo_class, mock_repo_instance


@pytest.fixture
def fetcher(tmp_path):
    """创建一个使用临时缓存目录的 RosMsgFetcher 实例。"""
    return RosMsgFetcher(cache_dir=tmp_path)


def test_fetch_package_clone(fetcher, mock_git_repo, tmp_path):
    """测试当缓存为空时，是否会克隆仓库。"""
    mock_repo_class, _ = mock_git_repo
    repo_url = ROS_MSG_REPOS["std_msgs"]
    package_path = fetcher.fetch_package("std_msgs", repo_url)

    # 验证 clone_from 被正确调用
    mock_repo_class.clone_from.assert_called_once_with(repo_url, tmp_path / "std_msgs")
    # 验证返回的路径是否正确
    assert package_path == tmp_path / "std_msgs" / "std_msgs"


def test_fetch_package_update(fetcher, mock_git_repo, tmp_path):
    """测试当仓库已存在时，是否会执行更新。"""
    mock_repo_class, mock_repo_instance = mock_git_repo
    repo_url = ROS_MSG_REPOS["std_msgs"]
    repo_path = tmp_path / "std_msgs"
    
    # 预先创建目录，模拟仓库已存在
    (repo_path / "std_msgs").mkdir(parents=True)

    fetcher.fetch_package("std_msgs", repo_url)

    # 验证是实例化的 Repo 对象，而不是 clone
    mock_repo_class.assert_called_with(repo_path)
    # 验证 pull 被调用
    mock_repo_instance.remotes.origin.pull.assert_called_once()


def test_find_and_fetch_known_package(fetcher, mock_git_repo):
    """测试查找一个已知的、有自己仓库的包。"""
    package_path = fetcher.find_and_fetch("std_msgs")
    assert package_path.name == "std_msgs"
    assert package_path.parent.name == "std_msgs"


def test_find_and_fetch_package_in_common_msgs(fetcher, mock_git_repo):
    """测试在 common_msgs 中查找一个包。"""
    package_path = fetcher.find_and_fetch("sensor_msgs")
    assert package_path.name == "sensor_msgs"
    assert package_path.parent.name == "common_msgs"


def test_find_and_fetch_package_not_found(fetcher, mock_git_repo):
    """测试当包在任何地方都找不到时，是否抛出 KeyError。"""
    with pytest.raises(KeyError, match="'non_existent_pkg' not found"):
        fetcher.find_and_fetch("non_existent_pkg")


def test_fetch_package_not_in_repo(fetcher, mock_git_repo):
    """测试当仓库中不包含请求的包时，是否抛出 FileNotFoundError。"""
    repo_url = ROS_MSG_REPOS["std_msgs"]
    with pytest.raises(FileNotFoundError, match="'wrong_pkg' not found in repository 'std_msgs'"):
        fetcher.fetch_package("wrong_pkg", repo_url)


from pathlib import Path
from collections import deque
from typing import List, Set

from .parser import MsgParser
from .generator import ProtoGenerator


class Converter:
    """The main class for converting ROS messages to Protobuf files."""

    def __init__(self, ros_distro: str = 'noetic'):
        # self._parser = MsgParser(ros_distro=ros_distro)
        self._parser = MsgParser()
        self._generator = ProtoGenerator()
        self._processed_messages = set()

    def convert(self, top_level_msg_type: str, output_dir: str):
        """
        Converts a top-level ROS message and its dependencies to .proto.

        Args:
            top_level_msg_type: The top-level message to convert (e.g., 'std_msgs/String').
            output_dir: The directory where .proto files will be saved.
        """
        output_path = Path(output_dir)
        queue = deque([top_level_msg_type])

        while queue:
            msg_type = queue.popleft()
            if msg_type in self._processed_messages:
                continue

            print(f"Processing {msg_type}...")
            try:
                package_name, msg_name = msg_type.split('/')
                parsed_msg = self._parser.parse(package_name, msg_name)

                proto_content, dependencies = self._generator.generate_proto(
                    parsed_msg,
                    package_name=package_name,
                    msg_name=msg_name
                )

                self._write_proto_file(output_path, package_name, msg_name, proto_content)

                for dep in dependencies:
                    if dep not in self._processed_messages:
                        queue.append(dep)

                self._processed_messages.add(msg_type)
                print(f"Successfully converted {msg_type}")

            except Exception as e:
                print(f"Failed to convert {msg_type}: {e}")
                # Re-raise the exception to halt the entire conversion process
                raise

    def _write_proto_file(self, output_dir: Path, package_name: str, msg_name: str, content: str):
        """Writes the .proto content to the appropriate file."""
        package_dir = output_dir / package_name
        package_dir.mkdir(parents=True, exist_ok=True)
        file_path = package_dir / f"{msg_name}.proto"
        file_path.write_text(content, encoding='utf-8')
        print(f"Wrote {file_path}")